from __future__ import annotations
from irys_sdk.fund import Fund
from irys_sdk.tokens.base import BaseToken
from irys_sdk.tokens.ethereum import EthereumToken
from irys_sdk.upload import Upload
from irys_sdk.bundle.tags import Tags
from irys_sdk.utils import get_balance, get_price


class Uploader:
    token_config: "BaseToken"
    url: str
    uploader: "Upload"
    token: str
    funder: "Fund"

    def __init__(self, url, token, **opts):
        self.url = url
        self.token = token
        self.uploader = Upload(self)
        self.funder = Fund(self)
        token_ref = get_token(token)
        token_opts = opts.get("token_opts")
        self.token_config = token_ref(self, **token_opts)
        self.token_config.ready()

    @property
    def address(self) -> str:
        return self.token_config.address

    def upload(self, data: bytearray, tags: Tags = None, target: str = None, anchor: str = None):
        return self.uploader.upload(data, tags, target, anchor)

    def get_balance(self) -> int:
        return get_balance(self.url, self.token, self.address)

    def get_price(self, bytes: int) -> int:
        return get_price(self.url, self.token, bytes)

    def fund(self, amount_atomic: int, multiplier=1.0):
        return self.funder.fund(amount_atomic, multiplier)


def get_token(token: str) -> "BaseToken":
    match token:
        case "ethereum":
            return EthereumToken
        case _:
            raise Exception("Unknown/unsupported token {}".format(token))



import codecs
from typing import Any
from irys_sdk.bundle.signers.signer import Signer
from irys_sdk.bundle.constants import SIG_CONFIG
from eth_account import Account
from eth_account.messages import encode_defunct, _hash_eip191_message
from eth_account._utils.signing import to_standard_v
from eth_keys import keys


class EthereumSigner(Signer):
    account: Account
    public_key = None
    private_key = None
    signature_type = 3
    signature_length = SIG_CONFIG[3]['sigLength']
    owner_length = SIG_CONFIG[3]['pubLength']

    def __init__(self, private_key: str):
        private_key = private_key[2:] if private_key.startswith(
            "0x") else private_key
        dec = codecs.decode(private_key, "hex")
        self.private_key = keys.PrivateKey(dec)
        self.public_key = b'\x04' + \
            self.private_key.public_key.to_bytes()

    def sign(self, message: bytearray, **opts: Any) -> bytearray:
        msg = encode_defunct(primitive=message)

        acc = Account.from_key(self.private_key)
        signature = (acc.sign_message(msg)).signature.hex()
        return bytearray.fromhex(signature[2:] if signature.startswith("0x") else signature)

    @staticmethod
    def verify(pubkey: bytearray, message: bytearray, signature: bytearray, **opts: Any) -> bool:
        msg = encode_defunct(primitive=message)
        msg_hash = _hash_eip191_message(msg)
        # trim pubkey
        pubkey = keys.PublicKey(
            pubkey if len(pubkey) == 64 else pubkey[1:])
        # standardize v value
        signature[64] = to_standard_v(signature[64])
        signature = keys.Signature(signature)
        valid = keys.ecdsa_verify(msg_hash, signature, pubkey)
        return valid




from urllib.parse import urljoin
import requests
from random import randbytes

from irys_sdk.bundle.tags import Tags
from irys_sdk.bundle.dataitem import DataItem
from irys_sdk.bundle.create import create_data
from irys_sdk.bundle.sign import sign


class Upload:
    irys: "Uploader"

    def __init__(self, irys: "Uploader"):
        self.irys = irys
        self.url = irys.url

    def upload_tx(self, tx: DataItem, **upload_opts):
        post_url = urljoin(self.url, "/tx/{}".format(self.irys.token))
        response = requests.post(post_url, data=tx.get_raw(), headers={
                                 "content-type": "application/octet-stream"})
        match response.status_code:
            case 429:
                raise Exception("Insufficient funds for the upload")
            case 200:
                return response.json()
            case _:
                raise Exception(
                    "Unexpected status code {} - {}".format(response.status_code, response.content))

    def upload(self, data: bytearray, tags: Tags = None, target: str = None, anchor: str = None, **upload_opts):
        signer = self.irys.token_config.get_signer()
        tx = create_data(data, signer, tags, target,
                         anchor if anchor else randbytes(16).hex())
        sign(tx, signer)
        if not tx.is_valid():
            raise Exception("internal error - produced data item is invalid")

        return self.upload_tx(tx, **upload_opts)



import base64
from irys_sdk.bundle.tags import encode_tags, Tags
from irys_sdk.bundle.dataitem import DataItem
from irys_sdk.bundle.signers.signer import Signer
from irys_sdk.bundle.utils import long_to_8_byte_array, set_bytes, short_to_2_byte_array


# type DataItemCreateOptions = {'target': Tags,
#                               'target': str, 'anchor': str}


def create_data(data: bytearray | str, signer: Signer, tags: Tags = None, target: str = None, anchor: str = None) -> DataItem:
    owner = signer.public_key
    # target = opts.get('target')
    target = None if target == None else base64.urlsafe_b64decode(target)
    target_length = 1 + (len(target) if target != None else 0)
    # anchor = opts.get('anchor')
    anchor = None if anchor == None else anchor.encode()
    anchor_length = 1 + (len(anchor) if anchor != None else 0)

    opt_tags = tags
    tags = None if opt_tags == None else encode_tags(opt_tags)
    tags_length = 16 + (0 if tags == None else len(tags))

    data = data.encode() if isinstance(data, str) else data

    data_length = len(data)

    length = 2 + signer.signature_length + signer.owner_length + \
        target_length + anchor_length + tags_length + data_length

    bytes = bytearray(length)

    set_bytes(bytes, short_to_2_byte_array(signer.signature_type), 0)

    set_bytes(bytes, bytearray(signer.signature_length), 2)

    if len(owner) != signer.owner_length:
        raise Exception("Owner must be {} bytes, but was incorrectly {} bytes".format(
            signer.owner_length, len(owner)))

    set_bytes(bytes, owner, 2 + signer.signature_length)

    position = 2 + signer.signature_length + signer.owner_length

    bytes[position] = 0 if target == None else 1
    if (target != None):
        if len(target) != 32:
            raise Exception(
                "Target must be 32 bytes, was incorrectly {} bytes".format(len(target)))
        set_bytes(bytes, target, position + 1)

    anchor_start = position + target_length
    tags_start = anchor_start + 1
    bytes[anchor_start] = 0 if anchor == None else 1
    if (anchor != None):
        tags_start += len(anchor)
        if (len(anchor) != 32):
            raise Exception(
                "Anchor must be 32 bytes, was incorrectly {} bytes".format(len(anchor)))
        set_bytes(bytes, anchor, anchor_start + 1)

    set_bytes(bytes, long_to_8_byte_array(
        0 if opt_tags == None else len(opt_tags)), tags_start)

    bytes_count = long_to_8_byte_array(0 if tags == None else len(tags))
    set_bytes(bytes, bytes_count, tags_start + 8)

    if (tags != None):
        set_bytes(bytes, tags, tags_start + 16)

    data_start = tags_start + tags_length

    set_bytes(bytes, data, data_start)

    return DataItem(bytes)



from abc import abstractmethod
from typing import Any


class Signer():

    @property
    @abstractmethod
    def public_key(self) -> bytearray:
        pass

    @property
    @abstractmethod
    def signature_type(self) -> int:
        pass

    @property
    @abstractmethod
    def signature_length(self) -> int:
        pass

    @property
    @abstractmethod
    def owner_length(self) -> int:
        pass

    @abstractmethod
    def sign(self, message: bytearray, **opts: Any) -> bytearray:
        pass

    @abstractmethod
    # @staticmethod
    def verify(pubkey: bytearray, message: bytearray, signature: bytearray, **opts: Any) -> bool:
        pass



from __future__ import annotations
import hashlib
from irys_sdk.bundle.dataitem import DataItem
from irys_sdk.bundle.signers.signer import Signer
from irys_sdk.bundle.utils import set_bytes


def sign(item: 'DataItem', signer: Signer) -> bytearray:
    (signature, id) = get_signature_and_id(item, signer)
    item.set_signature(signature)
    return id


def get_signature_and_id(item: 'DataItem', signer: Signer) -> (bytearray, bytearray):
    signature_data = get_signature_data(item)
    signature_bytes = signer.sign(signature_data)
    id_bytes = hashlib.sha256(signature_bytes).digest()
    return (signature_bytes,  id_bytes)


def get_signature_data(item: 'DataItem') -> bytearray:
    arr = [
        b"dataitem",
        b"1",
        str(item.signature_type).encode(),
        item.raw_owner,
        item.raw_target,
        item.raw_anchor,
        item.raw_tags,
        item.raw_data
    ]
    dh = deep_hash(arr)
    return dh


def deep_hash(data: list[bytearray]) -> bytearray:
    if type(data) == list:
        tag = b"list" + str(len(data)).encode()
        return deep_hash_chunks(data, hashlib.sha384(tag).digest())
    else:
        tag = b"blob" + str(len(data)).encode()
    tagged_hash = hashlib.sha384(tag).digest() + hashlib.sha384(data).digest()
    return hashlib.sha384(tagged_hash).digest()


def deep_hash_chunks(chunks, acc: bytearray):
    if len(chunks) < 1:
        return acc
    hash_pair = acc + deep_hash(chunks[0])
    new_acc = hashlib.sha384(hash_pair).digest()
    return deep_hash_chunks(chunks[1:], new_acc)
